#ifndef TrgReadout_HH
#define TrgReadout_HH

#include <iostream>
#include <fstream>

// hal
#include "VMEAddressTable.hh"
#include "VMEAddressTableASCIIReader.hh"
#include "DaqBusAdapter.hh"
//nclude "VMEDummyBusAdapter.hh"
#include "HardwareAccessException.hh"

// dual/inc/rcd
#include "RcdHeader.hh"
#include "RcdUserRW.hh"

#include "SubInserter.hh"
#include "SubAccessor.hh"

// online/inc/crc
#include "CrcVmeDevice.hh"
//#include "TrgReadoutStartupData.hh"
#include "TrgReadoutConfigurationData.hh"


class TrgReadout : public RcdUserRW {

public:
  TrgReadout(unsigned b, unsigned char c, unsigned char s) :
    _location(c,s,CrcLocation::beTrg,0),
    _crateNumber(c), _slotNumber(s), _device(0), _busAdapter(0),
    _addressTableReader("online/hal/CrcVmeAddress.hal"),
    _addressTable("CRC VME Address Table",_addressTableReader) {
    
    // Catch exceptions locally here in case PCI card not installed
    try {
      _busAdapter=new DaqBusAdapter(b);
    } catch ( std::exception e ) {
      _busAdapter=0;
    }
    
    if(_busAdapter!=0 && _slotNumber>=1 && _slotNumber<=21) {
      _device=new CrcVmeDevice(_addressTable,*_busAdapter,_slotNumber);
      if(_device->alive()) {
	std::cout << "TrgReadout::ctor()  PCI card " << b
		  << " slot " << std::setw(2)
		  << (unsigned)_slotNumber << " found alive" << std::endl;
      } else {
	delete _device;
	_device=0;
      }
    }
    
    // Allow no trigger for separate crates!
    //assert(_device!=0);

    // Define the poll objects
    _poll.location(_location);
    _pollSpill.location(_location);
  }
  
  virtual ~TrgReadout() {
    /* Causes a core dump???
    if(_busAdapter!=0) delete _busAdapter;
    if(_device!=0) delete _device;
    */
  }

  virtual bool record(RcdRecord &r) {
    if(doPrint(r.recordType())) {
      std::cout << "TrgReadout::record()" << std::endl;
      r.RcdHeader::print(std::cout," ") << std::endl;
    }

    // Catch exceptions from HAL code
    try {

      // Check record type
      switch (r.recordType()) {

      case RcdHeader::startUp: {
	if(_device!=0) assert(_device->beTrgSoftReset());
	break;
      }

      case RcdHeader::runStart: {

        // Access the DaqRunStart to get print level
        SubAccessor accessor(r);
        std::vector<const DaqRunStart*> v(accessor.extract<DaqRunStart>());
        assert(v.size()==1);
	_printLevel=v[0]->runType().printLevel();

	//assert(readRunData(r));

	break;
      }

      case RcdHeader::runEnd:
      case RcdHeader::runStop: {

	//assert(readRunData(r));

	break;
      }

	// Configuration start is used to set up system
      case RcdHeader::configurationStart: {

	// Zero counters, force busy, etc
	if(_device!=0) {
	  assert(_device->beTrgForceBusy(false));
	  assert(_device->beTrgSoftTrigger());
	  assert(_device->clearBeTrgFifos());
	}

	// Get the configuration information needed
	SubAccessor accessor(r);

	std::vector<const DaqConfigurationStart*>
	  dc(accessor.access<DaqConfigurationStart>());
	assert(dc.size()==1);
	_daqConfiguration=*(dc[0]);
	if(doPrint(r.recordType(),1)) _daqConfiguration.print(std::cout," ") << std::endl;
      
	std::vector<const TrgReadoutConfigurationData*>
	  bc(accessor.access<TrgReadoutConfigurationData>());
	assert(bc.size()<=1);
	if(bc.size()==1) _config=*(bc[0]);
	if(doPrint(r.recordType(),1)) _config.print(std::cout," ") << std::endl;
      
	/*
	std::vector<const CrcLocationData<CrcBeTrgConfigurationData>*>
	  bt(accessor.access< CrcLocationData<CrcBeTrgConfigurationData> >());

	if(doPrint(r.recordType(),1)) {
	  std::cout << " Number of CrcBeTrgConfigurationData subrecords = "
		    << bt.size() << std::endl;
	  for(unsigned i(0);i<bt.size();i++) {
	    bt[i]->print(std::cout,"  ") << std::endl;
	  }
	}
	assert(bt.size()<=3); // EMC, AHC slot broadcasts plus real trigger

	for(unsigned j(0);j<bt.size();j++) {
	  if(bt[j]->crateNumber()==_location.crateNumber()) {
	    assert(bt[j]->crcComponent()==CrcLocation::beTrg);
	    assert(bt[j]->label()==1);
	    
	    // Check if broadcast; if so, set for trigger slot
	    if(bt[j]->slotBroadcast()) {
	      if(doPrint(r.recordType(),1)) bt[j]->print(std::cout," ") << std::endl;
	      assert(_device->writeBeTrgConfigurationData(*bt[j]->data()));
	      
	      // If not broadcast, check slot numbers for match
	    } else {
	      assert(bt[j]->slotNumber()==_slotNumber);
	      if(doPrint(r.recordType(),1)) bt[j]->print(std::cout," ") << std::endl;
	      assert(_device->writeBeTrgConfigurationData(*bt[j]->data()));
	    }
	  }
	}

	assert(readConfigurationData(r));
	*/

	break;
      }


	// Configuration end reads back setup
      case RcdHeader::configurationEnd:
      case RcdHeader::configurationStop: {
	//assert(readConfigurationData(r));
	break;
      }

      case RcdHeader::acquisitionStart: {
	if(_config.enable()) {
	}

	_nEvents=0;
	break;
      }

      case RcdHeader::acquisitionEnd:
      case RcdHeader::acquisitionStop: {
	if(_config.enable()) {
	  
	  // Zero counters, force busy, etc
	  if(_device!=0) {
	    assert(_device->beTrgForceBusy(false));
	    assert(_device->beTrgSoftTrigger());
	    assert(_device->clearBeTrgFifos());
	  }
	}

	break;
      }

      case RcdHeader::spillStart: {
	if(_config.enable()) {
	  _spillStartTime=r.recordTime();
	  if(_device!=0 && _config.beTrgSoftSpill()) assert(_device->beTrgSoftSpill(true));
	}
	break;
      }

      case RcdHeader::spillEnd: {
	if(_config.enable()) {
	  if(_device!=0 && _config.beTrgSoftSpill()) assert(_device->beTrgSoftSpill(false));
	}
	break;
      }

      case RcdHeader::spill: {
	if(_config.enable()) {
	  *(_pollSpill.data())=TrgSpillPollData();
	  
	  if(_config.beTrgSoftSpill()) {
	    if(_printLevel>8) std::cout << "BeTrg soft spill on" << std::endl;
 
	    if(_device!=0) assert(_device->beTrgSoftSpill(true));
	    _pollSpill.data()->updateStartSpillTime();
	  }

	  if(_config.beTrgSoftTrigger()) {
	    if(_printLevel>8) std::cout << "BeTrg soft triggers" << _daqConfiguration.maximumNumberOfEventsInAcquisition() << std::endl;
	    if(_device!=0) assert(_device->beTrgSoftTriggers(_daqConfiguration.maximumNumberOfEventsInAcquisition()));

	  } else {
	    _pollSpill.data()->maximumSpillTime(_daqConfiguration.maximumTimeOfSpill());
	    _pollSpill.data()->maximumPollTime(_daqConfiguration.maximumTimeOfSpill());
	    _pollSpill.data()->maximumNumberOfEvents(_daqConfiguration.maximumNumberOfEventsInAcquisition());

	    if(_device!=0) assert(_device->trgSpillPoll(*(_pollSpill.data())));
	    if(doPrint(r.recordType(),1)) _pollSpill.print(std::cout," ");
	  
	    SubInserter inserter(r);
	    inserter.insert< CrcLocationData<TrgSpillPollData> >(_pollSpill);
	  }

	  if(_config.beTrgSoftSpill()) {
	    if(_printLevel>8) std::cout << "Be-Trg soft spill off" << std::endl;

	    if(_device!=0) assert(_device->beTrgSoftSpill(false));
	    _pollSpill.data()->updateEndSpillTime();
	  }

	  assert(takeTrigger(r));
	  assert(readEventData(r));
	}

	break;
      }

      case RcdHeader::trigger: {
	if(_config.enable()) {
	  assert(takeTrigger(r));
	  assert(readEventData(r));
	  _nEvents++;
	}
	break;
      }

      case RcdHeader::event: {
	if(_config.enable()) {
	  //assert(takeTrigger(r));
	  //assert(readEventData(r));
	  /*
	  unsigned char array[16*1024];
	  assert(_device->readVlinkEventData(*((CrcVlinkEventData*)array),true));
	  ((CrcVlinkEventData*)array)->print(std::cout,"TEST ",true);
	  */
	  
	}

	break;
      }

      case RcdHeader::postEvent: {

	break;
      }

      default: {

	break;
      }
      };

      return true;

    } catch ( HardwareAccessException& e ) {
      std::cout << "*** Exception occurred : " << e.what() << std::endl;
      
    } catch ( std::exception e ) {
      std::cout << "*** Unknown exception occurred" << std::endl;
    }
    return false;
  }

  bool readRunData(RcdRecord &r) {
    SubInserter inserter(r);

    CrcLocationData<CrcBeTrgRunData> *t(inserter.insert< CrcLocationData<CrcBeTrgRunData> >());
    t->location(_location);
    if(_device!=0) assert(_device->readBeTrgRunData(*t->data()));
    if(doPrint(r.recordType(),1)) t->print(std::cout," ") << std::endl;
    
    return true;
  }

  bool readConfigurationData(RcdRecord &r) {
    SubInserter inserter(r);

    CrcLocationData<CrcBeTrgConfigurationData> *b(inserter.insert< CrcLocationData<CrcBeTrgConfigurationData> >());
    b->location(_location);
    if(_device!=0) assert(_device->readBeTrgConfigurationData(*b->data()));
    if(doPrint(r.recordType(),1)) b->print(std::cout," ") << std::endl;
    
    return true;
  }

  bool takeTrigger(RcdRecord &r) {
    SubInserter inserter(r);
    
    if(_config.clearBeTrgTrigger()) {
      if(doPrint(r.recordType(),1)) std::cout << " Be-Trg clear trigger" << std::endl;
      if(_device!=0) assert(_device->clearBeTrgTrigger());
    }

    if(_config.beTrgSoftTrigger()) {
      if(doPrint(r.recordType(),1)) std::cout << " Be-Trg soft trigger" << std::endl;
      if(_device!=0) assert(_device->beTrgSoftTrigger());

    } else {
      _poll.data()->maximumTime(_daqConfiguration.maximumTimeOfEvent());
      UtlTimeDifference remainingTime(_daqConfiguration.maximumTimeOfSpill()-(r.recordTime()-_spillStartTime));
      if(remainingTime<_daqConfiguration.maximumTimeOfEvent()) _poll.data()->maximumTime(remainingTime);

      if(doPrint(r.recordType(),1)) _poll.print(std::cout," ") << std::endl;
      if(_device!=0) assert(_device->pollBeTrgTrigger(*(_poll.data())));
      if(doPrint(r.recordType(),1)) _poll.print(std::cout," ") << std::endl;

      inserter.insert< CrcLocationData<CrcBeTrgPollData> >(_poll);
    }

    return true;
  }

  bool readEventData(RcdRecord &r) {
    if(_config.readPeriod()>0 && (_nEvents%_config.readPeriod())==0) {
      SubInserter inserter(r);
      CrcLocationData<CrcBeTrgEventData>
	*p(inserter.insert< CrcLocationData<CrcBeTrgEventData> >());
      p->location(_location);

      //assert(_device->readBeTrgEventData(*p->data()));
      if(_device!=0) _device->readBeTrgEventData(*p->data(),_config.beTrgSquirt()); // Not assert because of FIFO
      inserter.extend(p->data()->numberOfFifoWords()*4); // For Be-Trg FIFO
      if(doPrint(r.recordType(),1)) p->print(std::cout," ") << std::endl;

    } else {
      if(_config.readcPeriod()>0 && (_nEvents%_config.readcPeriod())==0) {
	SubInserter inserter(r);
	CrcLocationData<CrcBeTrgEventData>
	  *p(inserter.insert< CrcLocationData<CrcBeTrgEventData> >());
	p->location(_location);
	
	if(_device!=0) assert(_device->readBeTrgEventData(*p->data(),false,true));
	if(doPrint(r.recordType(),1)) p->print(std::cout," ") << std::endl;
      }
    }

    return true;
  }
  

protected:
  const CrcLocation _location;
  const unsigned char _crateNumber;
  const unsigned char _slotNumber;

  CrcVmeDevice *_device;
  DaqBusAdapter *_busAdapter;
  VMEAddressTableASCIIReader _addressTableReader;
  VMEAddressTable _addressTable;

  DaqConfigurationStart _daqConfiguration;
  TrgReadoutConfigurationData _config;
  UtlTime _spillStartTime;

  CrcLocationData<CrcBeTrgPollData> _poll;
  CrcLocationData<TrgSpillPollData> _pollSpill;

  unsigned _nEvents;
};

#endif
