#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>

#include <iostream>
#include <sstream>
#include <vector>
#include <cstdio>

#include "VMEAddressTable.hh"
#include "VMEAddressTableASCIIReader.hh"
#include "SBS620x86LinuxBusAdapter.hh"
//#include "VMEDummyBusAdapter.hh"
#include "HardwareAccessException.hh"

#include "UtlTime.hh"
#include "UtlArguments.hh"
#include "RcdArena.hh"
#include "RcdWriterAsc.hh"
#include "RcdWriterBin.hh"
#include "RcdWriterDmy.hh"
#include "DaqRunStart.hh"

#include "SubAccessor.hh"

#include "CrcLocationData.hh"
#include "CrcReadout.hh"
#include "CrcReadoutConfigurationData.hh"
#include "OnlCounter.hh"
#include "HstChipNoise.hh"

using std::cin;
using std::exception;
using namespace std;

#define CERCVMEADDRESSTABLE "CercAddressMap.dat"
#define SEQUENCE_SETTINGS "Sequences.dat"


bool continueJob=true;
bool continueRun=true;
bool continueConfiguration=true;

void signalHandler(int signal) {
  std::cout << "Process " << getpid() << " received signal "
	    << signal << std::endl;
  continueJob=false;
  continueRun=false;
  continueConfiguration=false;
}

int main(int argc, const char **argv) {
  try{

    UtlArguments argh(argc,argv);
    //argh.print(cout);

    const bool doCheck(argh.option('c',"Run data check"));
    const bool doHistograms(argh.option('s',"Display histograms"));
    const bool doGraphics(argh.option('g',"Display graphics"));
    const bool useIntTrigger(argh.option('i',"Internal trigger selection"));
    
    const unsigned printLevel(argh.optionArgument('p',5,"Print level"));
    const unsigned feMask(argh.optionArgument('f',0xff,"FE mask"));
    const unsigned nEvents(argh.optionArgument('n',10,"Number of events between updates"));
    
    std::string logFile(argh.optionArgument('l',"log/msg.log","RS232 message log file"));
    
    if(argh.help()) return 0;
    
    if(doCheck) cout << "Data check selected" << endl;
    else           cout << "Data check not selected" << endl;
    if(doHistograms) cout << "Histograms display selected" << endl;
    else           cout << "Histograms display not selected" << endl;
    if(doGraphics) cout << "Graphics display selected" << endl;
    else           cout << "Graphics display not selected" << endl;
    if(useIntTrigger) cout << "Internal trigger selected" << endl;
    else              cout << "External trigger selected" << endl;
    
    cout << "Print level set to " << printLevel << endl;
    cout << "FE mask set to " << printHex((unsigned char)feMask) << endl;
    cout << "Number of events set to " << nEvents << endl;
    
    // Define the CERC locations
    unsigned char theCrate(0xec);
    //const unsigned nSlots(4);
    //unsigned theSlots[nSlots]={6,10,14,18};
    const unsigned nSlots(1);
    unsigned theSlots[nSlots]={18};
    
    SBS620x86LinuxBusAdapter busAdapter(0);
    VMEAddressTableASCIIReader addressTableReader( CERCVMEADDRESSTABLE );
    VMEAddressTable addressTable( "Test address table", addressTableReader );
    
    CrcVmeDevice *dev[nSlots];
    for(unsigned i(0);i<nSlots;i++) {
      dev[i]=new CrcVmeDevice(addressTable,busAdapter,theSlots[i]);
      assert(dev[i]->alive());
    }

    // Register CERCs
    CrcReadout oe(theCrate);
    oe.printLevel(printLevel);
    for(unsigned i(0);i<nSlots;i++) oe.device(dev[i],feMask,i==0);
    
    HstBase *hn(0);
    if(doHistograms) hn=new HstChipNoise(true);

    RcdArena arena;
    RcdArena *aArray(new RcdArena[256]);

    // Send Startup record
    OnlCounter od;
    od.readout(arena,RcdHeader::startUp);
    oe.record(arena);

    // Send Slow controls record
    od.readout(arena,RcdHeader::slowControl);
    SubInserter inserter(arena);

    for(unsigned i(0);i<nSlots;i++) {
      CrcLocationData<CrcAdm1025SlowControlsData>
	*b(inserter.insert< CrcLocationData<CrcAdm1025SlowControlsData> >());
      b->crateNumber(theCrate);
      b->slotNumber(theSlots[i]);
      b->crcComponent(CrcLocation::vme);
      b->label(1);

      *b->data()=CrcAdm1025SlowControlsData();
      b->print(std::cout) << std::endl;

      CrcLocationData<CrcLm82SlowControlsData>
	*l(inserter.insert< CrcLocationData<CrcLm82SlowControlsData> >());
      l->crateNumber(theCrate);
      l->slotNumber(theSlots[i]);
      l->crcComponent(CrcLocation::be);
      l->label(1);
      
      *l->data()=CrcLm82SlowControlsData();
      l->print(std::cout) << std::endl;
    }
  
    oe.record(arena);

    // Loop over runs
    for(unsigned iRun(0);iRun<1 && continueJob;iRun++) {

      // Send Run start
      od.readout(arena,RcdHeader::runStart);
      SubAccessor extracter(arena);
      std::vector<const DaqRunStart*> v(extracter.extract<DaqRunStart>());
      assert(v.size()==1);

      // Set run number
      unsigned runNumber=v[0]->runNumber();
      cout << "Run number = " << runNumber << endl;
      ostringstream oss;
      oss << "dat/Run" << runNumber;

      // Ignor Ctrl^C from now on
      signal(SIGINT,SIG_IGN);

      // Open output file
      RcdWriterAsc writer;
      //RcdWriterBin writer;
      //RcdWriterDmy writer;
      writer.open(oss.str());

      // Readout and write
      oe.record(arena);
      if(hn!=0) hn->record(arena);
      writer.write(&arena);
    
      // Loop over configurations
      //for(unsigned iConfiguration(0);iConfiguration<1024 && continueJob && continueRun;iConfiguration++) {
      for(unsigned iConfiguration(0);iConfiguration<32 && continueJob && continueRun;iConfiguration++) {
	unsigned dac0(0),dac1(0);
	//if((iConfiguration%2)==0) dac0=128*(511-(iConfiguration/2));
	//else                      dac1=128*(511-(iConfiguration/2));
	
	dac0=2048*(31-iConfiguration);
	dac1=0;

	cout << "DAC0 = " << dac0 << ", DAC1 = " << dac1 << endl;

	od.readout(arena,RcdHeader::configurationStart);

	SubInserter inserter(arena);

	// Add readout control
	CrcReadoutConfigurationData
	  *b(inserter.insert<CrcReadoutConfigurationData>());
	*b=CrcReadoutConfigurationData();
	b->vmePeriod(0);
	b->bePeriod(0);
	b->fePeriod(0);
	b->beTrgMode(1);
	b->vlinkMode(3);
	if(printLevel>1) b->print(std::cout) << std::endl;
	
	for(unsigned i(0);i<nSlots;i++) {
	  CrcLocationData<CrcBeConfigurationData>
	    *b(inserter.insert< CrcLocationData<CrcBeConfigurationData> >());
	  b->crateNumber(theCrate);
	  b->slotNumber(theSlots[i]);
	  b->crcComponent(CrcLocation::be);
	  b->label(1);

	  *b->data()=CrcBeConfigurationData();
	  b->data()->mode(0x02);
	  b->data()->test(0x00);
	  b->data()->testLength(0);
	  if(printLevel>2) b->print(std::cout) << std::endl;

	  // Only do BE-Trg for first board
	  if(i==0) {
	    CrcLocationData<CrcBeTrgConfigurationData>
	      *bt(inserter.insert< CrcLocationData<CrcBeTrgConfigurationData> >());
	    bt->crateNumber(theCrate);
	    bt->slotNumber(theSlots[i]);
	    bt->crcComponent(CrcLocation::beTrg);
	    bt->label(1);

	    *bt->data()=CrcBeTrgConfigurationData();
	    //	bt->data()->inputEnable(1<<24);
	    bt->data()->oscillationPeriod(40000); // 0.001 sec
	    if(printLevel>1) bt->print(std::cout) << std::endl;
	  }

	  for(unsigned f(0);f<8;f++) {
	    CrcLocationData<CrcFeConfigurationData>
	      *b(inserter.insert< CrcLocationData<CrcFeConfigurationData> >());
	    b->crateNumber(theCrate);
	    b->slotNumber(theSlots[i]);
	    b->crcComponent((CrcLocation::CrcComponent)f);
	    b->label(1);

	    *b->data()=CrcFeConfigurationData();
	    unsigned nadc=18;
	    b->data()->holdDuration(nadc*4*700);
	    b->data()->vfeMplexClockPulses(nadc);
	    b->data()->vfeTopGain(true);
	    b->data()->vfeBotGain(true);
	    b->data()->adcControl(8);// 8 = No DAC, 9 = DAC
	    b->data()->dacData(0,dac0);
	    b->data()->dacData(1,dac1);
	    if(printLevel>3) b->print(std::cout) << std::endl;

	    CrcLocationData<CrcFeFakeEventData>
	      *be(inserter.insert< CrcLocationData<CrcFeFakeEventData> >());
	    be->crateNumber(theCrate);
	    be->slotNumber(theSlots[i]);
	    be->crcComponent((CrcLocation::CrcComponent)f);
	    be->label(1);

	    be->data()->numberOfWords(32);
	    be->data()->enable(false);
	    unsigned *fe(be->data()->data());
	    for(unsigned ie(0);ie<be->data()->numberOfWords();ie++) {
	      if((ie%2)==0) fe[ie]=((~((   f)*4096+ie))<<16)&0xffff0000|((15-f)*4096+ie);
	      else          fe[ie]=((~((15-f)*4096+ie))<<16)&0xffff0000|((   f)*4096+ie);
	    }
	    inserter.extend(be->data()->numberOfWords()*4);

	    if(printLevel>3) be->print(std::cout) << std::endl;
	  }
	}

	oe.record(arena);
	if(hn!=0) hn->record(arena);
	writer.write(&arena);

	// Now catch Ctrl^C
	signal(SIGINT,signalHandler);

	// Loop over spills
	for(unsigned iSpill(0);iSpill<1 && continueJob && continueRun && continueConfiguration;iSpill++) {
	  od.readout(arena,RcdHeader::spillStart);
	  oe.record(arena);
	  if(hn!=0) hn->record(arena);
	  writer.write(&arena);

	  unsigned iEvent(0);
	  for(iEvent=0;iEvent<32 && continueConfiguration;iEvent++) {
	    //assert(dev[0]->clearBeTrgTrigger());
	    //assert(dev[0]->beTrgSoftTrigger());
	    //assert(dev[0]->beTrgSoftTrigger());

	    for(unsigned i(0);i<nSlots;i++) {
	      assert(dev[i]->clearBeTrgTrigger());
	      assert(dev[i]->beTrgSoftTrigger());
	      //assert(dev[i]->beTrgSoftTrigger());
	      //assert(dev[i]->beSoftTrigger()); //QWERTY
	      //assert(dev[i]->feSoftTrigger()); //QWERTY
	    }
	    
	    //assert(dev[0]->pollBeTrgTrigger(1000));
	    
	    od.readout(aArray[iEvent],RcdHeader::event);
	    oe.record(aArray[iEvent]);
	  }
      
	  if(continueConfiguration) od.readout(arena,RcdHeader::spillEnd);
	  else                      od.readout(arena,RcdHeader::spillStop);
	  oe.record(arena);

	  for(unsigned g(0);g<iEvent;g++) {
	    if(hn!=0) hn->record(aArray[g]);
	    writer.write(&aArray[g]);
	  }

	  if(hn!=0) {
	    hn->record(arena);
	    hn->update();
	    //if(iSpill>0) hn->postscript("junk");
	  }
	  writer.write(&arena);

	  od.readout(arena,RcdHeader::slowReadout);
	  oe.record(arena);
	  if(hn!=0) hn->record(arena);
	  writer.write(&arena);
	}

	signal(SIGINT,SIG_IGN);
	
	if(continueJob) od.readout(arena,RcdHeader::configurationEnd);
	else            od.readout(arena,RcdHeader::configurationStop);
	oe.record(arena);
	if(hn!=0) hn->record(arena);
	writer.write(&arena);
      }

      if(continueJob) od.readout(arena,RcdHeader::runEnd);
      else            od.readout(arena,RcdHeader::runStop);
      oe.record(arena);
      if(hn!=0) hn->record(arena);
      writer.write(&arena);

      writer.close();
    }

    if(hn!=0) delete hn;

  } catch ( HardwareAccessException& e ) {
    cout << "*** Exception occurred : " << e.what() << endl;
    
  } catch ( exception e ) {
    cout << "*** Unknown exception occurred" << endl;
  }
}
